/*
 * Copyright (c) 2001 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 */

import java.net.Socket;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.IOException;


public class Connection implements Runnable, TransferListener
{
	protected Socket		mClientSocket;
	protected Socket		mServerSocket;
	protected InputStream		mClientInputStream;
	protected OutputStream		mClientOutputStream;
	protected InputStream		mServerInputStream;
	protected OutputStream		mServerOutputStream;
	protected Sender		mSender;
	protected Receiver		mReceiver;
	protected String		mServerHostName;
	protected int			mServerPortNumber;
	protected int			mConnectionId;
	protected ConnectionListener	mListener;
	protected boolean		mServerDidConnection;
	protected Object		mServerLock = new Object();

	protected static int		mConnectionCounter;


	public Connection(Socket clientSocket)
	{
		mClientSocket = clientSocket;
		mConnectionId = ++mConnectionCounter;
	}


	/**
	 * Get the connection id for this connection.
	 * This is simply a monotonically increasing integer for each
	 * new connection for identification purposes.
	 */
	public int getId()
	{
		return mConnectionId;
	}


	public void setListener(ConnectionListener listener)
	{
		mListener = listener;
	}


	public ConnectionListener getListener()
	{
		return mListener;
	}


	/**
	 * Return the server host name which we are connecting to.
	 * This is always the final host specified in the URL, not
	 * that of the proxy that we actually connect to.  This is
	 * only available once the sender has initiated the connection.
	 */
	public String getServerHostName()
	{
		return mServerHostName;
	}


	/**
	 * Return the server port number which we are connectint to.
	 * This is always the final port number specified in the URL,
	 * not that of the proxy that we actually connect to.  This is
	 * only available once the sender has initiated the connection.
	 */
	public int getServerPortNumber()
	{
		return mServerPortNumber;
	}


	/**
	 * Return whether or not we have a connection to the remote server.
	 */
	public boolean isConnected()
	{
		return mServerSocket != null;
	}


	/**
	 * Method called to execute the Connection object in its own thread.
	 */
	public void run()
	{
		try
		{
			//
			// Tell any listeners that we are starting.
			//
			if (mListener != null)
				mListener.start(this);

			//
			// Get input and output streams for the socket.
			//
			mClientInputStream = mClientSocket.getInputStream();
			mClientOutputStream = mClientSocket.getOutputStream();

			//
			// Create the sender thread.
			//
			mSender = new Sender(this);
			mSender.setTransferListener(this);

			Thread senderThread = new Thread(mSender);

			senderThread.start();

			//
			// Create the receiver thread.
			//
			mReceiver = new Receiver(this);
			mReceiver.setTransferListener(this);

			Thread receiverThread = new Thread(mReceiver);

			receiverThread.start();

			//
			// Wait for the sending and receiving threads to exit.
			//
			senderThread.join();
			receiverThread.join();
		}
		catch (Exception e)
		{
			System.err.println("Connection " + mConnectionId +
				" failed: " + e);
		}

		//
		// Make sure that the socket is closed.
		//
		Utils.safeClose(mClientSocket);

		//
		// Tell any listeners that we are stopping.
		//
		if (mListener != null)
			mListener.stop(this);
	}


	/**
	 * Connect to the specified server and set up the input and
	 * output streams from the socket.  This is called by the sender
	 * after it has parsed the header and has figured out which
	 * server to connect to.  This notifies any waiting receiver
	 * that its receiving socket has been set up.
	 */
	public void connectToServer(String hostName, int portNumber)
	{
		//
		// Save the host name and port number being connected to.
		//
		mServerHostName = hostName;
		mServerPortNumber = portNumber;

		//
		// Inform any listener of the connection start.
		//
		if (mListener != null)
			mListener.connecting(this);

		//
		// Connect to the specified host.
		//
		try
		{
			//
			// Connect to the server.
			//
			mServerSocket = new Socket(hostName, portNumber);

			//
			// Save the input and output streams.
			//
			mServerInputStream = mServerSocket.getInputStream();
			mServerOutputStream = mServerSocket.getOutputStream();
		}
		catch (Exception e)
		{
			System.err.println("Connection failed: " + e);
			mServerSocket = null;
		}

		//
		// Tell the receiver if it was waiting that it can proceed.
		//
		notifyConnection();

		//
		// Inform any listener of the connection.
		//
		if (mListener != null)
			mListener.connected(this);
	}


	/**
	 * Notify the receiver that a server connection has been attempted.
	 * This is called whether or not the connection was successful.
	 * The isConnected method can be used to determine that fact.
	 */
	public void notifyConnection()
	{
		synchronized(mServerLock)
		{
			if (mServerDidConnection)
				return;

			//
			// We have finished doing the connection.
			//
			mServerDidConnection = true;

			//
			// Wake up the sender if necessary.
			//
			mServerLock.notify();
		}
	}


	/**
	 * Wait for a connection attempt to be made to the server.
	 * The connection is not necessary successful when this returns.
	 * The isConnected method can be called to determine this fact.
	 */
	public void waitForConnection()
	{
		synchronized(mServerLock)
		{
			while (!mServerDidConnection)
			{
				try
				{
					mServerLock.wait();
				}
				catch (InterruptedException e)
				{
				}
			}
		}
	}


	/**
	 * Return the input stream which reads from the client.
	 */
	public InputStream getClientInputStream()
	{
		return mClientInputStream;
	}


	/**
	 * Return the output stream which writes to the client.
	 */
	public OutputStream getClientOutputStream()
	{
		return mClientOutputStream;
	}


	/**
	 * Return the input stream which reads from the server.
	 */
	public InputStream getServerInputStream()
	{
		return mServerInputStream;
	}


	/**
	 * Return the output stream which writes to the Server.
	 */
	public OutputStream getServerOutputStream()
	{
		return mServerOutputStream;
	}


	/**
	 * Return the Sender for this connection.
	 */
	public Sender getSender()
	{
		return mSender;
	}


	/**
	 * Return the Receiver for this connection.
	 */
	public Receiver getReceiver()
	{
		return mReceiver;
	}


	/**
	 * Close down the connection.
	 * This can be called by either the Sender or the Receiver.
	 */
	protected void closeConnection()
	{
		//
		// Inform both the sender and the receiver that the
		// connection is being stopped.
		//
		if (mSender != null)
			mSender.stop();

		if (mReceiver != null)
			mReceiver.stop();

		//
		// Notify the receiver that the connection is finished
		// so that it will wake up if it was waiting there.
		//
		notifyConnection();

		//
		// Close the server and client sockets.
		//
		Utils.safeClose(mServerSocket);
		Utils.safeClose(mClientSocket);

		mServerSocket = null;
		mClientSocket = null;

		mClientInputStream = null;
		mClientOutputStream = null;
		mServerInputStream = null;
		mServerOutputStream = null;
	}


	//
	// ----------------------------------------------------------------
	// Implementation of the TransferListener interface.
	// This is used to inform our connection listeners of I/O progress.
	// ----------------------------------------------------------------
	//


	/**
	 * Method called when the first line of the data has been read.
	 * This will be the URL and METHOD line from a client, or the
	 * STATUS line from the server.
	 */
	public void firstLineAvailable(Transferer transferer)
	{
		if (mListener == null)
			return;

		if (transferer == mSender)
			mListener.requestFirstLine(this);

		if (transferer == mReceiver)
			mListener.responseFirstLine(this);
	}


	/**
	 * Method called when all of the header lines have been read.
	 */
	public void headersAvailable(Transferer transferer)
	{
		if (mListener == null)
			return;

		if (transferer == mSender)
			mListener.requestHeaders(this);

		if (transferer == mReceiver)
			mListener.responseHeaders(this);
	}


	/**
	 * Method called on every read operation.
	 */
	public void didRead(Transferer transferer)
	{
		if ((mListener != null) && (transferer == mReceiver))
			mListener.receive(this);
			
	}


	/**
	 * Method called on every write operation.
	 */
	public void didWrite(Transferer transferer)
	{
		if ((mListener != null) && (transferer == mSender))
			mListener.send(this);
	}
}
