/*
 * Copyright (c) 2001 David I. Bell
 * Permission is granted to use, distribute, or modify this source,
 * provided that this copyright notice remains intact.
 */

/**
 * An abstract class which manages multiple connection listeners.
 */
import java.util.Hashtable;


public abstract class ConnectionManager
{
	protected Hashtable connectionTable = new Hashtable(20);


	/**
	 * Initialize the connection manager in preparation for
	 * connection listeners.
	 */
	public void initialize()
	{
	}


	/**
	 * Assign a connection listener to be used for a new connection.
	 * This also adds the listener for the connection to our hash
	 * table so that we can manage it.
	 */
	public void assignConnectionListener(Connection connection)
	{
		//
		// Allocate the connection listener for the connection.
		//
		ConnectionListener listener = allocateConnectionListener();

		//
		// Put the listener into the hash table indexed by
		// the connection.
		//
		connectionTable.put(connection, listener);

		//
		// Tell the listener who we are so that it can inform us
		// of various events.
		//
		listener.setManager(this);

		//
		// Set the listener for the connection.
		//
		connection.setListener(listener);
	}


	/**
	 * Free the connection listener for a connection.
	 * This also removes the connection from our hash table so
	 * that it is no longer being managed.
	 */
	public void freeConnectionListener(Connection connection)
	{
		ConnectionListener listener = connection.getListener();

		if (listener == null)
			return;

		connection.setListener(null);

		connectionTable.remove(listener);
	}


	/**
	 * Inform the connection manager that one of its listeners has
	 * had new data.  This default method does nothing.
	 */
	public void changed(ConnectionListener listener)
	{
	}


	/**
	 * Allocate a connection listener for a new connection.
	 * This can either be a new listener used for a single
	 * connection or a shared listener which can handle
	 * multiple connections.
	 */
	protected abstract ConnectionListener allocateConnectionListener();
}
